//<c>   Copyright 1994-1998 by Gerry J. Danen; all rights reserved.

// Original SYNC is Copyright (c) 1992, Michael Holmes and Bob Flanders
// PC Magazine C Lab Notes, Synchronize two directories
// ********************************************************************

#include"gd_tools.h"
#include"gd_extrn.h"
#include<io.h>
#include<stdarg.h>                              // argument functions
#include<fcntl.h>                               // file control header
#include<stat.h>                                // file status header
#include<graph.h>                               // text mode routines


#define pgVERS_ "V4.01"
#define pgTITL_ "Synchronize Two Directories"
#define pgCOPR_ "1994"


#if     (! LARGE_MODEL)
  #error LARGE MEMORY MODEL REQUIRED
#endif
                                                // text window macros
#define gotoxy(x,y)     _settextposition(y, x)  // set cursor position
#define clrscr()        _clearscreen(_GWINDOW)  // clear screen
#define textcolor(x)    _settextcolor((short)x) // set up text and
#define textbackground(x) _setbkcolor(x)        // ..and bkgnd color
#define window(ur,uc,lr,lc) _settextwindow(uc,ur,lc,lr) // define logical window

                                                // shorthands
#define COPYBUF_SIZE 16384                      // copy buffer size
#define FILE_ATTR   _A_NORMAL | _A_RDONLY       // file attrib to search on
#define KEY_1       p->l_dir[0].d_compare       // compare key for dir 1
#define KEY_2       p->l_dir[1].d_compare       // ..and dir 2

                                                // key definitions
#define UP          0x100 + '\x48'              // up key
#define DOWN        0x100 + '\x50'              // down key
#define ESC         '\x1b'                      // escape key
#define F8          0x100 + '\x42'              // F8 function key
#define CTRL_LEFT   0x100 + '\x73'              // control left arrow key
#define CTRL_RIGHT  0x100 + '\x74'              // control right arrow key
#define LEFT        0x100 + 75                  // left arrow key
#define RIGHT       0x100 + 77                  // right arrow key
#define END         0x100 + '\x73'              // end key
#define PAGE_UP     0x100 + '\x49'              // page up key
#define PAGE_DOWN   0x100 + '\x51'              // page down key


// *********************************************
//      function definitions
// *********************************************

void    add_data(struct d_entry *d);            // copy in directory entry
void    add_file(int d);                        // add file to linked list
void interrupt far break_interrupt(void);       // control break interrupt
void    build_files_list(void);                 // build linked list
void    close_down(void);                       // close down full screen
void    color_set(enum select_types s_flag);    // select color palette
void    con_printf(char *msg,...);              // console output w/color
void    copy_a_file(int key);                   // copy one file
void far interrupt critical_handler(UINT bp,UINT di,UINT si,UINT ds,UINT es,UINT dx,UINT cx,UINT bx,UINT ax,UINT cs,UINT ip,UINT flags); // critical error handler
void    display_files_list(void);               // display linked list
void    display_line(struct lines *p,int line,enum select_types s); // display a directory pair
int     get_key(void);                          // get any type of key
void    home_screen(void);                      // display first screenful
void    sync_init(int ac,char *av[]);           // arg check and setup
int     key_hit(void);                          // key hit/^break check
void    move_screen(int key);                   // scroll screen one line
void    new_entry(struct lines *p,enum add_types bf,int index); // add a new entry to the linked list
void    pr_copyright(void);
void    quit_with(char *msg,...);               // quit with error message
enum set_up_rtns set_up_copy(int key,int *from,int *to); // determine source & dest directories for copy
void    set_up_screen(void);                    // set up base screen
void    sync_all_files(void);                   // synchronize all files
char *  translate_name(char *name);             // get true name
void    update_progress(void);                  // give progress info
char *  validate_name(char *name,char *rc);     // validate file/dir name
int     wherex(void);                           // return cursor column
int     wherey(void);                           // ..and cursor row


// ********************************************************************
//      globals
// ********************************************************************

int     rc = 1 ;                                // errorlevel return code
int     ctrl_break ;                            // control break flag
int     full_screen ;                           // full screen mode active
int     c_line ;                                // current line
int     text_fg = RED ;                         // default foreground color
int     text_bg = WHITE ;                       // ..background color
int     hdr_fg  = BR_WHITE ;                    // header foreground color
int     hdr_bg  = MAGENTA ;                     // ..background color
int     quit_fg = WHITE ;                       // when quitting
int     quit_bg = BLUE ;
int     max_lines ;                             // screen size
int     init_drive ;                            // initial drive setting
int     working_side = -1 ;                     // -1 is left, 1 is right

char    *args[2] ;                              // directory arguments
char    *paths[2] ;                             // base file paths
char    flags[2] ;                              // ..and file/dir flags
char    *buf ;                                  // disk read/write buffer
char    direction ;                             // copy direction indicator
char    *init_path ;                            // start up drive and path
char    whole_blk[17] ;                         // string of whole blocks
char    *part_blk = "\xdd" ;                    // ..string of a half blk
char    fullPath[2*FS_LEN] ;                    // path and filename for viewer

long    w_size ;                                // working file size
long    t_size ;                                // total file size


// ********************************************************************
//      structures
// ********************************************************************

enum    select_types                        // display line types
{
    selected ,                              // selected line
    unselected ,                            // normal display line
    border ,                                // border lines
    quitting ,                              // quitting
    selected_hilite                         // highlight selected line
};

enum    add_types                           // add entry positions
{
    before,                                 // before current entry
    after                                   // after current entry
};

enum    entry_types                         // directory entry types
{
    one_sided = ' ',                        // only one side used
    greater_than = '',                     // dir 1 > dir 2
    less_than = '',                        //       <
    equal_to = '',                         //       =
    almost_equal = '?'                      // file date only matches
};

enum    set_up_rtns                         // set_up_copy() returns
{
    set_up_ok,                              // set up completed ok
    no_file,                                // missing source file
    equal_files                             // equal timestamps
};

struct  d_entry                             // directory entry
{
    char d_name[13];                        // filename
    UINT d_date,                            // file date
         d_time;                            // ..time
    unsigned
    long d_size,                            // ..and size
         d_compare;                         // ..date/time compare key
    char d_attrib,                          // file attribute
         d_display[24];                     // size, date, time display
};

struct  lines                               // display lines linked list
{
    struct lines   *l_fwd, *l_back;         // forward and backward ptrs
    struct d_entry  l_dir[2];               // directory entries
    enum   entry_types l_entry;             // comparison char
}   *l_base,                                // anchor to linked list
    *l_top,                                 // ..current screen top
    *l_bottom,                              // ..current screen bottom
    *p;                                     // ..work pointer

union   REGS r;                             // cpu registers
struct  SREGS s;                            // segment registers
struct  find_t dos_entry;                   // DOS directory entry


// ********************************************************************
//      messages and strings
// ********************************************************************

char    copyright[]    = "SYNC " pgVERS_ " \xfe Copyright (c) 1992, Michael Holmes and Bob Flanders\t\n"
                         "\t     Copyright (c) " pgCOPR_"-"_THE_COPR_YEAR_", "_THE_COMPANY_NAME_"\t\t\t\t\n"
                         "\t     Synchronize two directories\n\n",
       *bad_drive[]  = { "Invalid drive %s\n",
                         "Drive not ready %s\n",
                         "Bad file specification %s\n" },
        empty_list[]   = "No files to process\n",
        only_dirs[]    = "%s is not a directory\n",
        same_path[]    = "Both arguments must not reference the same directory:\n1=%s\n2=%s\n",
        no_memory[]    = "Insufficient memory to continue processing\n",
        bad_width[]    = "Screen must be at least 80 columns wide\n",
        open_error[]   = "\nError opening file (%s)\n",
        read_error[]   = "\nError reading file (%s)\n",
        write_error[]  = "\nError writing file (%s)\n",
        stop_here[]    = "\nStopping at user's request\n",
        clear_error[]  = "\r%*s\r",
        ans_yes[]      = "Yes\r",
        ans_no[]       = "No\r",
        file_fmt[]     = "%8ld %02d/%02d/%2d %2d:%02d",
        line_fmt[]     = " %-12.12s%-24.24s",
        compare_fmt[]  = " %c%c ",
        header[]       = " %-39.39s%39.39s ",
        footer[]       = " Ctrl\x1a/\x1b to copy, F8 to sync, ESC to exit ",
        working[]      = "Copying: %-*.*s%-*.*s%*.*s\xb3 %c%c\r",
        done[]         = "SYNC completed normally\n",
        help[]         =
            "  Usage:  SYNC  source1 source2\n\n"
            "  Where:  source = d:\n"
            "                  [d:]path\n\n";

// ********************************************************************
//      print copyright notice
// ********************************************************************

void    pr_copyright( void )
{
    if ( ISLOADED_ANSI() )
        fprintf( stderr, "\033[2J\033[1;1H" );
    else
        fprintf( stderr, "\n\n\n" );
    printf( copyright );
}


// ********************************************************************
//      quit_with() -- give an error message, then return to DOS
// ********************************************************************

void    quit_with( char *msg, ... )             // string to print
{
    UINT    i;                                  // work parm for set drive
    va_list list;                               // variable list

    _dos_setdrive(init_drive, &i);              // reset the default drive
    chdir(init_path);                           // ..and the start up path

    if (full_screen)                            // q. full screen mode?
        close_down();                           // a. yes .. clear screen

    va_start(list, msg);                        // set up variable list
    vprintf(msg, list);                         // give error message ..

    gd_exit(rc,"");                             // ..and then quit
}


// ********************************************************************
//      build_files_list() -- build linked list of files to display
// ********************************************************************

void    build_files_list(void)
{
    int     i, j,                               // loop controls
            rc;                                 // find return code
    char    w_path[FS_LEN];                     // path and files mask

    printf( "Collecting file data.  Please wait..." );
    for (i = 0; i < 2; i++)                     // for both subdirectories
    {
        sprintf(w_path, "%s%s*.*",              // build findfirst mask
                paths[i],
                ST_LASTCHAR(paths[i]) == '\\' ? "" : "\\");

        rc = _dos_findfirst(w_path, FILE_ATTR, &dos_entry); // find first requested file

        for (j = 0; ! rc; j++)                  // find all requested files
        {
            add_file(i);                        // add entry to list
            rc = _dos_findnext(&dos_entry);     // ..then try to find next
        }
    }

    if (! l_base)                               // q. any files found?
        quit_with(empty_list);                  // a. no .. give error/quit

    for (p = l_base; p; p = p->l_fwd)           // run thru linked list
    {
        if (! p->l_dir[0].d_name[0] ||          // q. only one side used..
                ! p->l_dir[1].d_name[0])        // ..by a dir entry?
            p->l_entry = one_sided ;            // a. yes .. set to blank
        else
        if ( KEY_1 == KEY_2 )                   // q. same date and time?
        {                                       // a. yes .. now check size
            if (p->l_dir[0].d_size ==           // q. both sizes same?
                        p->l_dir[1].d_size)
                p->l_entry = equal_to;          // a. yes .. set equal flag
            else
                p->l_entry = almost_equal;      // else .. show almost same
        }
        else
        if (KEY_1 > KEY_2)                      // q. dir 1 > dir 2?
            p->l_entry = greater_than;          // a. yes .. set .gt. flag
        else
            p->l_entry = less_than;             // else .. set less than flg
    }
}


void    erase_file( int side )
{
    // side 0 is left, 1 is right
    if ( p->l_dir[side].d_name[0] != EOS &&
         p->l_dir[side].d_name[0] != ' ' )
    {
        UCHAR Ques[FS_LEN*2], *save_line = NULPTR ;
        sprintf( Ques, "Delete %s", p->l_dir[side].d_name );
        Ques[68] = EOS ;
        sn_beep();
        save_line = SaveLine( vid.rows, save_line );
        if ( YesNo( Ques, NO, YES ) )           // ask for confirmation
        {
            sprintf( fullPath, "%s%s", paths[side], p->l_dir[side].d_name ); // build fully qualified name
            fl_delete( fullPath );
//~~~
//  Fix display
            p->l_dir[side].d_name[0] = ' ' ;
            p->l_dir[side].d_name[1] = p->l_dir[side].d_display[0] = EOS ;
//     ???  display_line(p, c_line, selected);      // update line display
        }
        save_line = RestoreLine( vid.rows, save_line );
    }
}


void    view_file( int side )
{
    // side 0 is left, 1 is right
    if ( p->l_dir[side].d_name[0] != EOS &&
         p->l_dir[side].d_name[0] != ' ' )
    {
        sprintf( fullPath, "%s%s", paths[side], p->l_dir[side].d_name ); // build fully qualified name
        fl_view_init( sc_vid_rows()-2, sc_vid_cols()/2+1, 0, "", YES ); // init viewer
        fl_view( fullPath, FALSE );
        fl_view_exit();
    }
}


// ********************************************************************
//      display_files_list() -- display file linked list
// ********************************************************************

void    display_files_list( void )
{
    int key, i ;

    set_up_screen();                            // set up base screen
    home_screen();                              // display initial lines

    for (;;)                                    // loop 'til user quits
    {
        key = get_key();                        // get keystroke
        switch ( key )                          // see what we got
        {
            case 0:                             // no key available
                continue;                       // ..try again later

            case F8:                            // F8 function key
                sync_all_files();               // syncronize all files
                close_down();                   // clean up screen
                return;                         // ..and return to caller

            case PAGE_UP:                       // page up key
                move_screen(UP);                // move the first line..

                for (i = max_lines - 3; i--;)   // ..then loop for the rest
                {
                    if (p->l_back)              // q. line available?
                        move_screen(UP);        // a. yes .. scroll one line
                     else
                        break;                  // else .. exit loop
                }
                break;

            case PAGE_DOWN:                     // up key
                move_screen(DOWN);              // move the first line..

                for (i = max_lines - 3; i--;)   // ..then loop for the rest
                {
                    if (p->l_fwd)               // q. line available?
                        move_screen(DOWN);      // a. yes .. scroll one line
                     else
                        break;                  // else .. exit loop
                }
                break;

            case UP:                            // down key
            case DOWN:                          // up key
                move_screen(key);               // handle scrolling
                break;

            case CTRL_RIGHT:                    // ctrl right arrow
            case CTRL_LEFT:                     // ctrl left arrow
                copy_a_file( key );             // copy one file
                break;

            case LEFT :                         // left arrow
                working_side = -1 ;
                display_line(p, c_line, selected); // redisplay current line
                break;
            case RIGHT :                        // right arrow
                working_side = 1 ;
                display_line(p, c_line, selected); // redisplay current line
                break;

            case 'd':                           // delete a file
            case 'D':
                if ( working_side < 0 )
                    erase_file( 0 );            // delete leftside file
                else
                    erase_file( 1 );            // delete rightside file
                break ;

            case 'v':                           // view a file
            case 'V':
                if ( working_side < 0 )
                    view_file( 0 );             // show leftside file
                else
                    view_file( 1 );             // show rightside file
                break ;

            case ESC:                           // escape key
                close_down();                   // clean up screen..
                return;                         // ..and exit gracefully

            default:                            // error case
                sn_beep();
        }
    }
}


// ********************************************************************
//      display_line() -- display one directory pair
// ********************************************************************

void    display_line(struct lines *p,           // line to display
                     int  line,                 // ..on which line
                     enum select_types s)       // selected/unselected flag
{
    gotoxy(1, line);                            // move to proper line
    color_set(unselected);                      // use unselected palette
    con_printf( s ? " " : "" );                // ..put out spacer char

    if ( s==selected && working_side < 0 )      // choose colors
        color_set(selected_hilite);
    else
        color_set(s);

    con_printf(line_fmt, p->l_dir[0].d_name,    // display left side
                p->l_dir[0].d_display);

    color_set(border);                          // choose border colors

    switch( p->l_entry )                        // show file relationship
    {
        case greater_than :
            con_printf( compare_fmt, '', p->l_entry );
            break ;
        case less_than :
            con_printf( compare_fmt, p->l_entry, '' );
            break ;
        default :
            con_printf( compare_fmt, p->l_entry, p->l_entry );
            break ;
    }

    if ( s==selected && working_side > 0 )      // choose colors again
        color_set(selected_hilite);
    else
        color_set(s);

    con_printf(line_fmt, p->l_dir[1].d_name,    // display right side
                p->l_dir[1].d_display);

    color_set(unselected);                      // use unselected palette
    con_printf( s ? " " : "" );                // ..put out other side

    gotoxy(1, line);                            // position at start of line
}


// ********************************************************************
//      add_file() -- add file entry to linked list
// ********************************************************************

void    add_file( int d )                       // directory number
{
    int     rc;                                 // string compare rtn code

    for (p = l_base; p; )                       // look down list of entries
    {
        if (! d)                                // q. working on first dir?
        {                                       // a. no .. do right side
            if ((strcmp(dos_entry.name,         // q. find where to..
                    p->l_dir[0].d_name)) < 0)   // ..insert entry into list?
            {
                new_entry(p, before, 0);        // a. yes .. before current
                return;                         // ..then return to caller
            }
        }
        else
        {
            if (p->l_dir[0].d_name[0])          // q. left entry active?
            {                                   // a. yes .. process it
                if (! (rc = strcmp(             // q. find a matching..
                    dos_entry.name,             // ..file in the other dir?
                    p->l_dir[0].d_name)))
                {
                    add_data(&p->l_dir[1]);     // a. yes .. copy in entry
                    return;                     // ..and return to caller
                }
            }
            else
            {
                rc = strcmp(                    // else .. check and see..
                dos_entry.name,                 // ..where this entry..
                p->l_dir[1].d_name);            // ..belongs in the list
            }

            if (rc < 0)                         // q. fit in here?
            {
                new_entry(p, before, d);        // a. yes .. add in before
                return;                         // ..and return to caller
            }
        }

        if (p->l_fwd)                           // q. find end of list?
            p = p->l_fwd;                       // a. no .. continue looping

        else
        {
            new_entry(p, after, d);             // else .. add to end
            return;                             // ..and return to caller
        }
    }

    new_entry(p, after, d);                     // add the 1st entry to list
}


// ********************************************************************
//      new_entry() -- add a new entry to linked list
// ********************************************************************

void    new_entry( struct lines *p,             // base entry
                   enum add_types bf,           // before/after flag
                   int index )                  // index number for entry
{
    struct  lines *n;                           // work pointer

    if ((n = (struct lines *) malloc(           // q. get storage for a..
             sizeof (struct lines))) == 0)      // ..new line entry?
        quit_with(no_memory);                   // a. no .. give error/quit

    memset(n, 0, sizeof(struct lines));         // clear entry to nulls

    if (! p)                                  // q. adding first entry?
    {
        l_base = n;                             // a. yes .. set up base ptr
        n->l_fwd = n->l_back = 0;               // ..and clr both pointers
    }
    else
    {
        if (bf == before)                       // q. before current entry?
        {                                       // a. yes .. chain into list
            if (p->l_back)                      // q. previous entry exist?
            {
                (p->l_back)->l_fwd = n;         // a. yes .. patch fwd ptr
                n->l_back = p->l_back;          // ..and new backward ptr
            }
            else
            {
                l_base = n;                     // else .. new one is first
                n->l_back = 0;                  // ..and no backward pointer
            }

            n->l_fwd = p;                       // set up new's forward ptr
            p->l_back = n;                      // and next's backward ptr
        }
        else
        {
            if (p->l_fwd)                       // q. next entry exist?
            {
                (p->l_fwd)->l_back = n;         // a. yes .. patch to next..
                n->l_fwd = p->l_fwd;            // ..entry fwd and backward
            }
            else
                n->l_fwd = 0;                   // ..and no more afterwards

            n->l_back = p;                      // set up new's backward ptr
            p->l_fwd = n;                       // and next's backward ptr
        }
    }

    add_data(&n->l_dir[index]);                 // finally copy in dir entry
}


// ********************************************************************
//      add_data() -- add directory information to list entry
// ********************************************************************

void    add_data(struct d_entry *d)             // directory entry
{
    strncpy((char *) d->d_name,                 // pick up filename
                dos_entry.name, 13);
    d->d_attrib = dos_entry.attrib;             // ..and attribute
    d->d_date = dos_entry.wr_date;              // .. file date
    d->d_time = dos_entry.wr_time;              // ..time
    d->d_size = dos_entry.size;                 // ..and size

    d->d_compare = ((long) dos_entry.wr_date    // build compare key..
                   << 16) + dos_entry.wr_time;  // ..from date and time

    sprintf(d->d_display, file_fmt, d->d_size,  // format size/date/time..
                (d->d_date >> 5) & 0xf,         // ..for screen display
                d->d_date & 0x1f, ((d->d_date >> 9) & 0x7f)+ 80,
                (d->d_time >> 11) & 0x1f, (d->d_time >> 5) & 0x3f);
}


// ********************************************************************
//      sync_all_files() -- synchronize all files in list
// ********************************************************************

void    sync_all_files( void )
{
    home_screen();                              // display from top of list
    for (;;)                                    // loop thru linked list..
    {
        copy_a_file( F8 );                      // copy file..
        if (p->l_fwd)                           // q. more lines available?
            move_screen(DOWN);                  // a. yes .. move one line
         else
            break;                              // else .. done loop
    }
}


// ********************************************************************
//      copy_a_file() -- copy one file
// ********************************************************************

void    copy_a_file( int key )                  // key hit
{
    int     from, to,                           // src/dest indices/handles
            b_read;                             // bytes read
    char    from_name[FS_LEN],                  // from path and filename
            to_name[FS_LEN];                    // to path and filename
    enum    set_up_rtns rc;                     // set up return code
    struct  d_entry *f;                         // "from" directory entry


    if ((rc = set_up_copy(key, &from, &to))     // q. request without a ..
                == no_file)                     // ..source file?
    {
        sn_beep();
        return;                                 // ..so just return now
    }
    else
    if (rc == equal_files)                      // q. files equal and sync
        return;                                 // a. yes .. just return

    window(1, 1, 80, max_lines);                // define window coordinates
    color_set(border);                          // set up border colors
    gotoxy(1, max_lines);                       // position at bottom

//  if ( check_file_dates() )                   // q. continue copy?
//  {                                           // a. yes .. process file
        direction = (from == 0) ?               // copy direction indicator
                greater_than : less_than;       // ..for status display

        f = &p->l_dir[from];                    // set up ptr to dir entry

        memcpy(&p->l_dir[to], f,                // update directory entry..
                sizeof(struct d_entry));        // ..for "to" size of fence

        strcpy(from_name, paths[from]);         // build fully qualified ..
        strcat(from_name, f->d_name);           // .."from" filename

        strcpy(to_name, paths[to]);             // build fully qualified ..
        strcat(to_name, f->d_name);             // .."to" filename

        if ((from = open(from_name,             // q. open input ok?
                O_BINARY | O_RDONLY | S_IREAD)) == -1)
        {
            ErrMsg( "Cannot open file %s for input", from_name );
            goto CleanUp ;
        }

        fl_delete( to_name );                   // delete output file first

        if ((to = open(to_name, (O_CREAT | O_WRONLY | O_BINARY | O_TRUNC), S_IWRITE)) == -1)
        {                                       // handle error
            fl_delete(to_name);                 // delete file
            if ((to = open(to_name, (O_CREAT | O_WRONLY | O_BINARY | O_TRUNC), S_IWRITE)) == -1) // try again
            {
                ErrMsg( "Cannot open file %s for output", to_name );
                goto CleanUp ;
            }
        }

        for (t_size = w_size = f->d_size ; ((b_read = read(from, buf, COPYBUF_SIZE)) != -1) ; )
        {
            if (! b_read)                       // q. at end of file?
                break;                          // a. yes .. exit loop

            if (b_read != write(to, buf, b_read))   // q. write ok?
                quit_with(write_error, to_name);    // a. no .. give error

            w_size -= b_read;                   // subtract out amount read
            update_progress();                  // keep user informed
        }

        if (b_read == -1 || w_size)             // q. err or didn't finish?
            quit_with(read_error, from_name);   // a. yes .. give error/quit

        _dos_setftime(to,                       // set file's date and time
                f->d_date, f->d_time);

        _dos_setfileattr(to_name,               // ..and file's attributes
                f->d_attrib);

        close(to);                              // close output file
        close(from);                            // ..and input file

        p->l_entry = equal_to;                  // show both files equal
//  }

CleanUp:
    con_printf(clear_error,                     // fix up display by
            81 - sizeof(footer), "");           // ..clearing message area
    window(1, 2, 80, max_lines - 1);            // redefine text window
    display_line(p, c_line, selected);          // update line display
}


// ********************************************************************
//      check_file_dates() -- check for invalid file dates
// ********************************************************************

// int     check_file_dates(void)
// {
//     int     key;                                // work and key variable
//
//     if (! ((p->l_dir[0].d_name[0] &&            // q. dir 1 in use..
//             ! (p->l_dir[0].d_date & 0xfc00))    // ..and year 1980
//             || ((p->l_dir[1].d_name[0] &&       // ..or dir 2 in use..
//             ! (p->l_dir[1].d_date & 0xfc00))))) // ..and year 1980
//         return(1);                              // a. no .. just return
//
//     con_printf(bad_date);                       // give notice of bad date
//
//     while(! (key = getch()))                    // wait for a key to be hit
//         ;
//
//     if (key == ESC)                             // q. user request cancel?
//         quit_with(stop_here);                   // a. yes .. give error/quit
//
//     key = (key == 'Y' || key == 'y' ||          // change response to..
//             key == '\r');                       // ..a logical (true/false)
//
//     con_printf(key ? ans_yes : ans_no);         // show user's response
//     tm_sleep_secs(1);                           // wait a second
//     return(key);                                // ..then return w/response
// }


// ********************************************************************
//      set_up_copy() -- determine source and destination files
// ********************************************************************

enum    set_up_rtns set_up_copy(int  key,       // key hit
                                int *from,      // source directory index
                                int *to)        // destination dir index
{
    int name_index;                             // dir entry with name

    name_index = p->l_dir[0].d_name[0] ? 0 : 1; // dir entry with filename

    switch (key)
    {
        case CTRL_RIGHT:
            if (p->l_entry == one_sided &&      // q. only in one dir..
                    name_index)                 // ..and not in dir 1?
                return(no_file);                // a. yes .. return w/error

            *from = 0;                          // source is dir 1 ..
            *to = 1;                            // ..and dest is dir 2
            break;                              // ..continue with process

        case CTRL_LEFT:
            if (p->l_entry == one_sided &&      // q. only in one dir..
                    ! name_index)               // ..and not in dir 2?
                return(no_file);                // a. yes .. return w/error

            *from = 1;                          // source is dir 2 ..
            *to = 0;                            // ..and dest is dir 1
            break;                              // ..continue with process

        case F8:                                // F8 sync request
            switch (p->l_entry)                 // based on comparison char
            {
                case equal_to:                  // both files equal
                case almost_equal:              // ..or almost equal
                    return(equal_files);        // ..no processing needed
                case greater_than:              // dir 1 greater than dir 2
                    *from = 0;                  // from file is in dir 1
                    *to = 1;                    // ..and to file is in dir 2
                    break;                      // ..continue with process
                case less_than:                 // dir 1 is less than dir 2
                    *from = 1;                  // from file is in dir 2
                    *to = 0;                    // ..and to file is in dir 1
                    break;                      // ..continue with process
                case one_sided:                 // file only in one dir
                    *from = name_index;         // source entry has filename
                    *to = name_index ? 0 : 1;   // dest is the other entry
                    break;                      // ..continue with process
            }
            break;                              // ..continue some more
    }
    return(set_up_ok);                          // rtn w/everything ok flag
}



// ********************************************************************
//      update_progress() -- update progress bar
// ********************************************************************

void    update_progress(void)
{
    static  int last_call = 0xff;               // value at last display
    int     half,                               // half unit flag
            skip;                               // skip space
    long    whole;                              // work number

    whole = ((t_size - w_size) << 5) / t_size;  // scale up by # of units
                                                // ..then divide to get # of
    half = (int) whole & 1;                     // ..units and half units completed so far

    if ( last_call != whole )                   // any change in graph?
    {
        last_call = (int) whole ;               // save status

        whole >>= 1;                            // get whole block count
        skip = 16 - (int) whole - half;         // spaces to skip in bar

        con_printf(working, (int) whole,        // display crude graph..
                (int) whole,  whole_blk,        // ..using whole blocks..
                half, half, part_blk,           // ..and partial units..
                skip, skip, "",                 // ..and skip to end of bar
                direction, direction);          // ..& draw a direction flag
    }
}



// ********************************************************************
//      home_screen() -- display first group of lines
// ********************************************************************

void    home_screen(void)
{
    int i;                                      // loop counter

    for (i = 0, l_top = l_bottom = p = l_base;  // loop thru building..
                i < (max_lines - 2); i++)       // ..initial screen
    {
        display_line(l_bottom, i + 1,           // ..and display one line
                i ? unselected : selected);

        if (l_bottom->l_fwd)                    // q. something follows?
            l_bottom = l_bottom->l_fwd;         // a. yes .. set up next one
         else
            break;                              // else .. exit loop
    }

    if (i == (max_lines - 2))                   // q. fill the screen?
        l_bottom = l_bottom->l_back;            // a. yes .. get last ptr

    gotoxy(1, c_line = 1);                      // position on first line
}


// ********************************************************************
//      move_screen() -- move scroll bar and redisplay lines
// ********************************************************************

void    move_screen(int key)                    // key code
{
    switch (key)
    {
        case UP:                                // up key
            display_line(p, c_line, unselected);// unselect line
            if (p == l_top)                     // q. cursor on top row?
            {                                   // a. yes .. try moving up
                if (p->l_back)                  // q. any previous lines?
                {
                    _scrolltextwindow(-1);      // a. yes .. make room
                    l_top = p = p->l_back;      // get top line's pointer
                    l_bottom = l_bottom->l_back;// ..and bottom line's ptr
                }
                else
                    sn_beep();                  // beep
            }
            else
            {
                p = p->l_back;                  // get previous line pointer
                c_line--;                       // move back one line
            }
            display_line(p, c_line, selected);  // write new current line
            break;                              // ..and continue process

        case DOWN:                              // down key
            display_line(p, c_line, unselected);// unselect line
            if (p == l_bottom)                  // q. cursor on bottom row?
            {                                   // a. yes .. try moving dwn
                if (p->l_fwd)                   // q. any more lines?
                {                               // a. yes .. scroll window
                    gotoxy(1, 1);               // goto top line ..
                    _scrolltextwindow(1);       // delete top line
                    l_top = l_top->l_fwd;       // get new top line pointer
                    p = l_bottom = p->l_fwd;    // ..and bottom line's ptr
                    gotoxy(1, max_lines - 2);   // set up shop at bottom
                }
                else
                    sn_beep();                  // beep
            }
            else
            {
                p = p->l_fwd;                   // get next line pointer
                c_line++;                       // ..and move down one line
            }
            display_line(p, c_line, selected);  // write new current line
            break;                              // ..and continue process
    }
}


// ********************************************************************
//      set_up_screen() -- set up screen, headers and footers
// ********************************************************************

void    set_up_screen(void)
{
    full_screen = 1;                            // set up full screen flag
    _settextrows(max_lines);                    // ..and max screen lines

    window(1, 1, 80, max_lines);                // define window coordinates
    color_set(unselected);                      // set up standard colors
    clrscr();                                   // clear screen

    _wrapon(_GWRAPOFF);                         // disable EOL scrolling
    gotoxy(1, 1);                               // goto top of screen
    color_set(border);                          // set up border colors
    con_printf( header, paths[0], paths[1] );   // ..display header

    gotoxy(82 - sizeof(footer), max_lines);     // position at bottom right
    con_printf(footer);                         // ..and display footer

    gotoxy(1, max_lines);                       // goto bottom left corner
    con_printf(clear_error,                     // fix up error area by
            81 - sizeof(footer), "");           // ..clearing to blanks

    window(1, 2, 80, max_lines - 1);            // define text window
}


// ********************************************************************
//      color_set() -- select current foreground and background colors
// ********************************************************************

void    color_set( enum select_types s_flag )   // selected/unselected
{
    switch (s_flag)                             // based on flag ..
    {
        case selected:                          // selected line
            textcolor(text_bg);                 // set up the inverted..
            textbackground(text_fg);            // ..color scheme
            break;                              // ..then continue

        case selected_hilite :                  // selected line highlighted
            textcolor(text_bg+8);               // set up the color scheme
            textbackground(text_fg);
            break;

        case unselected:                        // unselected line
            textcolor(text_fg);                 // set up the normal..
            textbackground(text_bg);            // ..color scheme
            break;                              // ..then continue

        case border:                            // border lines
            textcolor(hdr_fg);                  // set up the border..
            textbackground(hdr_bg);             // ..color scheme
            break;                              // ..then continue

        case quitting :                         // quitting time
            textcolor(quit_fg);
            textbackground(quit_bg);
            break;
    }
}


// ********************************************************************
//      close_down() -- close windows and quit to DOS
// ********************************************************************

void    close_down(void)
{
    window(1, 1, 80, max_lines);                // define full screen window
    color_set( quitting );                      // set up normal colors
    clrscr();                                   // clear screen
    full_screen = 0;                            // ..and full screen flag
    pr_copyright();
}


// ********************************************************************
//      get_key() -- get a key (including function keys)
// ********************************************************************

int     get_key(void)
{
    int k = 0 ;                                 // local key variable

    if (kbhit())                                // q. key available?
    {                                           // a. yes .. process it
        if (! (k = getch()))                    // q. function key?
            k = 0x100 + getch();                // a. yes .. special key
    }
    return( k );                                // return w/key if available
}


// ********************************************************************
//      validate_name() -- validate file/dir name
// ********************************************************************

char   *validate_name(char *name,           // source file/dir name
                      char *rc)             // return code
                                            //  -3 = invalid drive
                                            //  -2 = drive not ready
                                            //  -1 = invalid name
                                            //   0 = directory
                                            //   1 = filespec
{
    int i,                                  // work variable
        drive;                              // drive number
    static
    char    full[FS_LEN];                       // normalized full name
    char    *p,                                 // work pointer
            c,                                  // character holder
            work[FS_LEN],                       // work field
            current_path[FS_LEN];               // current path


    full[0] = 0;                                // initialize return string
    memset(work, 0, sizeof(work));              // clear to nulls
    strcpy(work, name);                         // copy source local
    i = strlen(work) - 1;                       // ..and pick up length

    if (work[i] == '\\' && i                    // q. more than drive:\?
                && (i > 2 || work[1] != ':'))   // ..or other trailing \?
        work[i] = 0;                            // a. yes .. remove last \ .

    drive = (work[1] == ':') ?                  // get name's drive number
                 toupper(work[0]) - 'A' + 1 : 0;

    if ((_osmajor == 3 && _osminor >= 1) ||     // q. DOS 3.1 or higher?
                (_osmajor > 3))
    {
        r.x.ax = 0x4409;                        // a. yes .. do local test
        r.h.bl = drive;                         // bl = drive to test
        int86(0x21, &r, &r);                    // test drive validity

        if (r.x.cflag)                          // q. bad drive?
        {
            *rc = -3;                           // a. yes .. set return code
            return(0);                          // ..and return to caller
        }
    }

    if (! (_getdcwd(drive, current_path,      // q. accessible drive?
                FS_LEN)))
    {
        *rc = -2;                               // a. no .. set error return
        return(0);                              // ..and return to caller
    }

    *rc = -1;                                   // set up for error return

    if (chdir(work))                            // q. a directory?
    {                                           // a. no .. backup one \ .
        if ((p = strrchr(work, '\\')) != 0)     // q. find a backslash?
        {                                       // a. yes .. check it out
            if ((p - work) > 2 ||               // q. more than just d:\?
                        work[1] != ':' ||       // ..or no colon anyway..
                        work[0] == '.')         // ..or some .\ or ..\?
            {
                *p++ = 0;                       // a. yes .. backoff filename
                c = 0;                          // ..and clear char holder
            }
            else
            {
                c = *(++p);                     // get 1st char after \ .
                *p++ = 0;                       // ..and make into new path
            }

            if (! chdir(work))                // q. find the directory?
            {
                *rc = 1;                        // a. yes .. show filename
                _getdcwd(drive, full, FS_LEN);  // ..then get path

                if (ST_LASTCHAR(full) != '\\')  // q. last char a \?
                    strcat(full, "\\");         // a. no .. put in path sep

                if ( c )                        // q. filespec need help?
                    strncat(full, &c, 1);       // a. yes .. add saved char

                strcat(full, p);                // ..and add filename
            }
        }
        else
        {
            _getdcwd(drive, full, FS_LEN);      // else .. then get path

            p = &work[work[1] == ':' ? 2 : 0];  // get start of filespec
            i = strlen(p);                      // ..and length of remainder

            if (ST_LASTCHAR(full) != '\\' && i) // q. do we need a \?
                strcat(full, "\\");             // a. yes .. put in path sep

            if (i)                              // q. filename given?
            {
                *rc = 1;                        // a. yes .. return filespec
                strcat(full, p);                // ..and copy in filename
            }
            else
                *rc = 0;                        // else .. show as dir
        }
    }
    else
    {
        *rc = 0;                                // show directory was found
        _getdcwd(drive, full, FS_LEN);          // ..then get path
    }

    chdir(current_path);                        // reset original directory

    return(*rc >= 0 ? strdup(full) : 0);        // rtn normalized path\name
}


// ********************************************************************
//      translate_name() -- translate a DOS directory name
// ********************************************************************

char    *translate_name(char *name)             // name to translate
{
    static
    char    work[65];                           // work/return area
    char   *wp = work;                          // ..and a pointer

    if (_osmajor < 3)                           // q. translate available?
       return(name);                            // a. no .. rtn with source

    r.h.ah = 0x60;                              // ah = translate
    r.x.si = FP_OFF(name);                      // set ptr to input name
    s.ds = FP_SEG(name);                        // ..and segment
    r.x.di = FP_OFF(wp);                        // set ptr to output area
    s.es = FP_SEG(wp);                          // ..and segment
    int86x(0x21, &r, &r, &s);                   // translate the name

    return(r.x.cflag ? NULL : work);            // return workarea or null
}


// ********************************************************************
//      critical_handler() -- DOS critical error handler
// ********************************************************************

void    far interrupt critical_handler(UINT bp,
            UINT di, UINT si, UINT ds, UINT es,
            UINT dx, UINT cx, UINT bx, UINT ax,
            UINT cs, UINT ip, UINT flags)
{
    if (ax & 0x800)                     // q. fail allowed?
        ax = (ax & 0xff00) | 3;         // a. yes .. fail request
    else
        ax = (ax & 0xff00) | 2;         // else .. abort request
}


// ********************************************************************
//      break_interrupt() -- control break interrupt routine
// ********************************************************************

void    interrupt far break_interrupt(void)
{
    ctrl_break = 1;                             // ^c or ^break entered
}


// ********************************************************************
//      key_hit() -- key hit function w/control break check
// ********************************************************************

int     key_hit(void)
{
    if (ctrl_break)                     // q. requested to stop?
        quit_with(stop_here);           // a. yes .. give error/quit
    return(kbhit());                    // else .. get keyboard stat
}


// ********************************************************************
//      wherex() -- determine column number
// ********************************************************************

int     wherex(void)
{
    struct rccoord pos;                 // row and col coordinates

    pos = _gettextposition();           // get current cursor ..
    return(pos.col);                    // ..and return the column
}


// ********************************************************************
//      wherey() -- determine row number
// ********************************************************************

int     wherey(void)
{
    struct rccoord pos;                 // row and col coordinates

    pos = _gettextposition();           // get current cursor ..
    return(pos.row);                    // ..and return the row nbr
}


// ********************************************************************
//      con_printf() -- console print routine for color output
// ********************************************************************

void    con_printf( char *msg, ... )    // string to print
{
    static va_list list;                // variable list

    va_start(list, msg);                // set up variable list
    vsprintf(buf, msg, list);           // format into buffer
    _outtext(buf);                      // ..and output to screen
}


// ********************************************************************
//      sync_init() -- init interrupts and parse command line
// ********************************************************************

void    sync_init( int ac, char *av[] )
{
    int     i;                                  // loop count
    char    w_path[FS_LEN],                     // working path name
            *t_path[2];                         // translated path
    struct  videoconfig vc;                     // screen info structure

    if ( ac != 3 || ! strcmp(av[1], "/?") )     // need help?
        quit_with(help);

    _dos_getdrive((UINT *) &init_drive);        // get the default drive
    init_path = getcwd(NULL, FS_LEN);           // ..and current directory

    _dos_setvect(0x23, break_interrupt);        // set up ctrl break handler
    _dos_setvect(0x24, critical_handler);       // ..DOS critical handlers

    _getvideoconfig(&vc);                       // get current screen info
    max_lines = vc.numtextrows;                 // save maximum nbr of lines

    if (vc.numtextcols < 80)                    // q. less than 80 columns?
        quit_with(bad_width);                   // a. yes .. give error/quit

    if (vc.mode == _TEXTBW80 || vc.mode == _TEXTMONO)   // no color
    {
        quit_bg = text_bg = hdr_bg = BLACK;
        quit_fg = text_fg = hdr_fg = WHITE;
    }

    for (i = 0; i < 2; i++)                     // for each directory..
    {
        if (! (args[i] = validate_name(         // q. valid file/directory?
                av[i + 1], &flags[i])))
            quit_with(bad_drive[flags[i]+3],    // a. no .. give error msg
                av[i + 1]);                     // ..and exit back to DOS

        if (flags[i])                           // q. argument a filename?
            quit_with(only_dirs, args[i]);      // a. yes .. give error/quit

        strcpy(w_path, args[i]);                // copy in base path

        if (! flags[i])                         // q. directory name?
        {                                       // a. yes .. check format
            if (! (ST_LASTCHAR(w_path) == '\\'))   // q. dir end in backslash?
                strcat(w_path, "\\");           // a. yes .. add the *.*
        }
        else
            *(strrchr(w_path, '\\') + 1) = EOS; // end path after last \ .

        paths[i] = strdup(w_path);              // save a copy of path
        t_path[i]= strdup(translate_name(w_path)); // and translate pathname for later

        if (ST_LASTCHAR(t_path[i]) == '\\')     // q. last char a backslash?
            ST_LASTCHAR(t_path[i]) = EOS ;      // a. yes .. remove it
    }

    if ( st_eq( t_path[0], t_path[1] ) )        // same path?
    {
        printf( "\n1-%s\n2-%s\n\n", av[0], av[1] );
        quit_with( same_path, t_path[0], t_path[1] );
    }

    for ( i = 2 ; i ; )                         // for each translated name
        free( t_path[--i] );                    //  free temporary memory

    if ((buf = malloc(COPYBUF_SIZE)) == 0)      // enough memory?
        quit_with(no_memory);                   // no: abort

    memset(whole_blk, 0xdb, 16);                // string of whole blocks for copy progress graph
}


void    main( int argc, char *argv[] )
{
    gd_init( 0 );

    pr_copyright();
    sync_init(argc, argv);                      // init and parse cmd line
    build_files_list();                         // build list of files
    display_files_list();                       // ..then display them

    rc = 0;                                     // clear errorlevel and
    quit_with(done);                            // ..return to DOS
}
